/*======================================================================

  $Id: ftl.c,v 1.2 1999/01/08 16:54:45 dwmw2 Exp $

    A Flash Translation Layer memory card driver

    This driver implements a disk-like block device driver with an
    apparent block size of 512 bytes for flash memory cards.

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#undef PSYCHO_DEBUG 

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>

#include "ftl.h"
#include "mtd.h"

#define FTL_MAJOR 44

/* Funky stuff for setting up a block device */
#define MAJOR_NR		FTL_MAJOR
#define DEVICE_NAME		"FTL"
#define DEVICE_REQUEST		do_ftl_request
#define DEVICE_ON(device)
#define DEVICE_OFF(device)

#define DEVICE_NR(minor)	((minor)>>3)
#define REGION_NR(minor)	((minor)&7)
#define MINOR_NR(dev,rgn)	(((dev)<<3)+(rgn))

#include <linux/blk.h>


#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
static char *version =
"ftl_cs.c 1.29 1997/06/01 05:03:16 (David Hinds)";
#else
#define DEBUG(n, args...)
#endif
#define MAX_DEV 0xffff
#define printd(args...) 

#ifndef printd
#define printd(args...) printk(args)
#endif

static int shuffle_freq = 50;

/*====================================================================*/

/* Sector size -- shouldn't need to change */
#define SECTOR_SIZE	512


/* Block IO request handler stuff */
void do_ftl_request(void);

/* Each memory region corresponds to a minor device */
typedef struct partition_t {
	struct mtd_info *mtd;
	u_long		state;
	u_long		*VirtualBlockMap;
	u_long		*VirtualPageMap;
	u_long		FreeTotal;
	struct eun_info_t {
		u_long			Offset;
		u_long			EraseCount;
		u_long			Free;
		u_long			Deleted;
	} *EUNInfo;
	struct xfer_info_t {
		u_long			Offset;
		u_long			EraseCount;
		u_short			state;
	} *XferInfo;
	u_short		bam_index;
	u_long		*bam_cache;
	u_short		DataUnits;
	u_long		BlocksPerUnit;
	erase_unit_header_t	header;
	int			open;
	struct wait_queue *wq;
} partition_t;

static partition_t myparts[MAX_MTD_DEVICES];


static void ftl_notify_func(struct mtd_info *mtd, int device);
void ftl_freepart(partition_t *part);

static struct mtd_notifier ftl_notifier={ftl_notify_func, NULL};

/* Partition state flags */
#define FTL_FORMATTED	0x01

/* Transfer unit states */
#define XFER_UNKNOWN	0x00
#define XFER_ERASING	0x01
#define XFER_ERASED	0x02
#define XFER_PREPARED	0x03
#define XFER_FAILED	0x04


int ftl_devsizes[256] = { 0. /* ... */ };
int ftl_blocksizes[256] ={ 0, /* ... */ };
int ftl_partsizes[256] ={ 0, /* ... */ };


struct hd_struct part_table[256];



void dummy_init(struct gendisk *crap)
{
}



struct gendisk ftl_gendisk={
        FTL_MAJOR,      /* Major number */      
        "ftl",          /* Major name */
        4,              /* Bits to shift to get real from partition */
        15,             /* Number of partitions per real */
        MAX_MTD_DEVICES,/* maximum number of real */
        dummy_init,     /* init function */
        part_table,     /* hd struct */
        ftl_partsizes,  /* block sizes */
        0,              /* number */
        NULL,           /* internal use, not presently used */
        NULL            /* next */
};


/*====================================================================*/

int ftl_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg);
int ftl_open(struct inode *inode, struct file *file);
int ftl_close(struct inode *inode, struct file *file);



struct file_operations ftl_blk_fops = {
    NULL,		/* lseek */
    block_read,		/* read */
    block_write,	/* write */
    NULL,		/* readdir */
    NULL,		/* select */
    ftl_ioctl,		/* ioctl */
    NULL,		/* mmap */
    ftl_open,		/* open */
    NULL,		/* flush */
    ftl_close,		/* release */
    block_fsync		/* fsync */
};

/*====================================================================*/


void ftl_erase_callback(struct erase_info *done)
{
	partition_t *part = (partition_t *)done->priv;
	struct xfer_info_t *xfer;
	int i;
	
	/* Look up the transfer unit */

	for (i = 0; i < part->header.NumTransferUnits; i++)
		if (part->XferInfo[i].Offset == done->addr) break;
	if (i == part->header.NumTransferUnits) {
		printk(KERN_NOTICE "ftl: internal error: "
		       "erase lookup failed!\n");
		return;
	}
	xfer = &part->XferInfo[i];
	if (done->state == MTD_ERASE_DONE)
		xfer->state = XFER_ERASED;
	else {
		xfer->state = XFER_FAILED;
		printk(KERN_NOTICE "ftl: erase failed: state = %d\n",
		       done->state);
	}

	wake_up(&part->wq);
	kfree(done);
}



/*======================================================================

    Scan_partition() checks to see if a memory region contains an FTL
    partition.  It reads all the erase unit headers, builds the erase
    unit map, and then builds the virtual page map.
    
======================================================================*/

int scan_partition(partition_t *part)
{
	erase_unit_header_t *header;
	u_long offset;
	u_long windowlen;
	u_short xvalid, xtrans, i;
	u_long blocks, j;
	int hdr_ok;
	int firstlogEUN;
	int ret;
	
	/* Search first megabyte for a valid FTL header */
	
	for (offset = 0;
	     offset < 0x100000;
	     offset += part->mtd->erasesize)
	{
		ret = MTD_POINT(part->mtd, offset, sizeof(erase_unit_header_t),
				&windowlen, (u_char **) (&header));
		if (ret)
			return ret;
		if (windowlen < sizeof(erase_unit_header_t))
			printk(KERN_NOTICE "ftl: %lx-byte window not big enough for FTL header.\n", windowlen);
		
		if (strcmp(header->DataOrgTuple+3, "FTL100") == 0) break;
		MTD_UNPOINT (part->mtd, offset);
	}

	if (offset == 0x100000) {
		printk(KERN_NOTICE "ftl: FTL header not found.\n");
		return -ENODEV;
	}
	
	if ((header->NumEraseUnits > 65536) || (header->BlockSize != 9) ||
	    (header->EraseUnitSize < 10) || (header->EraseUnitSize > 31) ||
	    (header->NumTransferUnits >= header->NumEraseUnits)) {

		printk(KERN_NOTICE "ftl: FTL header corrupt!\n");
		return -EIO;
	}
	
	memcpy (&part->header, header, sizeof(*header));
	
	if ((firstlogEUN=header->LogicalEUN))
		printk (KERN_NOTICE "ftl_cs: First Logical EUN is %d, not zero. Compensating...\n",firstlogEUN);
	
	
	/* Set up erase unit maps */
	part->DataUnits = header->NumEraseUnits - header->NumTransferUnits;
	part->EUNInfo = kmalloc(part->DataUnits * sizeof(struct eun_info_t),
				GFP_KERNEL);
	for (i = 0; i < part->DataUnits; i++)
		part->EUNInfo[i].Offset = 0xffffffff;
	part->XferInfo =
		kmalloc(header->NumTransferUnits * sizeof(struct xfer_info_t),
			GFP_KERNEL);
	
	xvalid = xtrans = 0;

	MTD_UNPOINT (part->mtd,offset);

	for (i = 0; i < part->header.NumEraseUnits; i++) 
	{
		offset = ((i + part->header.FirstPhysicalEUN)
			  << part->header.EraseUnitSize);
		ret = MTD_POINT(part->mtd, offset, sizeof(erase_unit_header_t),
				&windowlen, &header);
		
		if (ret) 
		{
			printk("MTD_POINT failed(1)\n");
			return ret;
		}
		
		/* Is this a transfer partition? */
		hdr_ok = (strcmp(header->DataOrgTuple+3, "FTL100") == 0);
		if (hdr_ok && (header->LogicalEUN-firstlogEUN < part->DataUnits) &&
		    (part->EUNInfo[header->LogicalEUN-firstlogEUN].Offset == 0xffffffff)) 
		{
			part->EUNInfo[header->LogicalEUN-firstlogEUN].Offset = offset;
			part->EUNInfo[header->LogicalEUN-firstlogEUN].EraseCount =
				header->EraseCount;
			xvalid++;
		}
		else {
			if (xtrans == part->header.NumTransferUnits) {
				printk(KERN_NOTICE "ftl_cs: format error: too many "
				       "transfer units (looking for %d)!\n",part->header.NumTransferUnits);
				MTD_UNPOINT (part->mtd, offset);
				return -EIO;
			}
			if (hdr_ok && (header->LogicalEUN == 0xffff)) {
				part->XferInfo[xtrans].state = XFER_PREPARED;
				part->XferInfo[xtrans].EraseCount = header->EraseCount;
			}
			else {
				part->XferInfo[xtrans].state = XFER_UNKNOWN;
				/* Pick anything reasonable for the erase count */
				part->XferInfo[xtrans].EraseCount =
					part->header.EraseCount;
			}
			part->XferInfo[xtrans].Offset = offset;
			xtrans++;
		}
		MTD_UNPOINT(part->mtd, offset);
	}
/* Check for format trouble */
	header = &part->header;
	if ((xtrans != header->NumTransferUnits) ||
	    (xvalid+xtrans != header->NumEraseUnits)) {
		printk(KERN_NOTICE "ftl_cs: format error: erase units "
		       "don't add up!\n");
		return -EIO;
	}
	
	/* Set up virtual page map */
	blocks = header->FormattedSize >> header->BlockSize;
	part->VirtualBlockMap = vmalloc(blocks * sizeof(u_long));
	memset(part->VirtualBlockMap, 0xff, blocks * sizeof(u_long));
	part->BlocksPerUnit = (1 << header->EraseUnitSize) >> header->BlockSize;
	
	
	part->bam_cache = kmalloc(part->BlocksPerUnit * sizeof(u_long),
				  GFP_KERNEL);
	part->bam_index = 0xffff;
	part->FreeTotal = 0;
	for (i = 0; i < part->DataUnits; i++) {
		part->EUNInfo[i].Free = 0;
		part->EUNInfo[i].Deleted = 0;
		offset = part->EUNInfo[i].Offset + part->header.BAMOffset;
#ifdef ASDASD	
if (!i)		printk("Trying to point to %lx for BAM %x, %lx bytes\n",offset,i, sizeof(u_long) * part->BlocksPerUnit );	
#endif

		ret = MTD_READ(part->mtd, offset, (u_long) (part->BlocksPerUnit * sizeof(u_long)),
				&windowlen, (u_char *)part->bam_cache, 0);

		if (ret) {
			printk("MTD_READ of BAM %d failed\n",i);
			return ret;
		}

		for (j = 0; j < part->BlocksPerUnit; j++) 
		{
#ifdef ASDASD
				if (0)printk("Checking block %lx, BAM %lx: %lx - ",j,i,part->bam_cache[j]);
#endif
			if (BLOCK_FREE(part->bam_cache[j])) 
			{
#ifdef ASDASD
				if (0)printk("Free");
#endif
				part->EUNInfo[i].Free++;
				part->FreeTotal++;
			}
			else if ((BLOCK_TYPE(part->bam_cache[j]) == BLOCK_DATA) &&
				 (BLOCK_NUMBER(part->bam_cache[j]) < blocks))
			{
#ifdef ASDASD
				if (0)printk("VBM");
#endif				
				part->VirtualBlockMap[BLOCK_NUMBER(part->bam_cache[j])] =
					(i << part->header.EraseUnitSize) + (j << part->header.BlockSize);
			}
			else if (BLOCK_DELETED(part->bam_cache[j]))
			{
#ifdef ASDASD
				if (0)printk ("Deleted");
#endif
				part->EUNInfo[i].Deleted++;
			}
#ifdef ASDASD
			if (0)printk("\n");
#endif
		}
	}
	printk("Valid FTL partition found.\n");
	return 0;
	
} /* scan_partition */


/*======================================================================

    Erase_xfer() schedules an asynchronous erase operation for a
    transfer unit.
    
======================================================================*/

static int erase_xfer(partition_t *part, u_short xfernum)
{
    int ret;
    struct xfer_info_t *xfer;
    struct erase_info *erase;

/*    printk("erase_xfer(%d)\n",xfernum);
 */   xfer = &part->XferInfo[xfernum];
    DEBUG(1, "ftl_cs: erasing xfer unit at 0x%lx\n", xfer->Offset);
    xfer->state = XFER_ERASING;
    /* Is there a free erase slot?  Always. */

    erase=kmalloc(sizeof(struct erase_info), GFP_KERNEL);
    if (!erase) 
	    return -ENOMEM;
    
    
    erase->callback=ftl_erase_callback;
    erase->addr = xfer->Offset;
    erase->len = part->mtd->erasesize;
    erase->priv = (u_long)part;
  
    ret = MTD_ERASE(part->mtd, erase); 

    if (!ret) 
	    xfer->EraseCount++;
    return ret;
} /* erase_xfer */

/*======================================================================

    Prepare_xfer() takes a freshly erased transfer unit and gives
    it an appropriate header.
    
======================================================================*/

static void prepare_xfer(partition_t *part, int i)
{
    erase_unit_header_t header;
    struct xfer_info_t *xfer;
    int nbam, ret;
    u_long ctl,retlen;
    u_long offset;
/*
    printk("prepare_xfer(%d)\n",i);
    */
    xfer = &part->XferInfo[i];
    xfer->state = XFER_FAILED;
    
    DEBUG(1, "ftl_cs: preparing xfer unit at 0x%lx\n", xfer->Offset);

    /* Write the transfer unit header */
    header = part->header;
    header.LogicalEUN = 0xffff;
    header.EraseCount = xfer->EraseCount;
    offset = xfer->Offset;

    ret = MTD_WRITE(part->mtd, xfer->Offset, sizeof(header), &retlen, (u_char *)&header, 0);
    if (ret)
    {
	    printk("mtd_write() failed in prepare_xfer!\n");
	    return;
    }

    /* Write the BAM stub */
    nbam = (part->BlocksPerUnit * sizeof(u_long) +
	    part->header.BAMOffset + SECTOR_SIZE - 1) / SECTOR_SIZE;
    ctl = BLOCK_CONTROL;
    offset += part->header.BAMOffset;

    for (i = 0; i < nbam; i++, offset += sizeof(u_long)) {
	    
	    ret = MTD_WRITE(part->mtd, offset, sizeof(u_long), &retlen, (u_char *)&ctl, 0);
	    if (ret)
	    {
		    printk("mtd_write() failed in prepare_xfer!\n");
		    return;
	    }
	    xfer->state = XFER_PREPARED;
	    
    } /* prepare_xfer */
}

/*======================================================================

    Copy_erase_unit() takes a full erase block and a transfer unit,
    copies everything to the transfer unit, then swaps the block
    pointers.

    All data blocks are copied to the corresponding blocks in the
    target unit, so the virtual block map does not need to be
    updated.
    
======================================================================*/

static int copy_erase_unit(partition_t *part, u_short srcunit,
			   u_short xferunit)
{
    u_char buf[SECTOR_SIZE];
    struct eun_info_t *eun;
    struct xfer_info_t *xfer;
    u_long offset, retlen;
    u_long src, dest, free, i;
    u_short unit;
    int ret;

    eun = &part->EUNInfo[srcunit];
    xfer = &part->XferInfo[xferunit];
    DEBUG(2, "ftl_cs: copying block 0x%lx to 0x%lx\n",
	  eun->Offset, xfer->Offset);
	
    /* Read current BAM */
    if (part->bam_index != srcunit) {
	offset = eun->Offset + part->header.BAMOffset;
	
	ret = MTD_READ(part->mtd, offset, part->BlocksPerUnit * sizeof(u_long),
			&retlen, (u_char *) (part->bam_cache), 0);

	/* mark the cache bad, in case we get an error later */
	part->bam_index = 0xffff;

	if (ret)
	{
		printk("ftl: Failed to read BAM cache in copy_erase_unit()!\n");
		
		return ret;
	}
    }
    
    /* Write the LogicalEUN for the transfer unit */
    xfer->state = XFER_UNKNOWN;


    offset = xfer->Offset + 20; /* Bad! */
    unit = 0x7fff;
    
    ret = MTD_READ(part->mtd, offset, sizeof(u_short),
			&retlen, (u_char *) &unit, 0);

	/* mark the cache bad, in case we get an error later */
	part->bam_index = 0xffff;

	if (ret)
	{
		printk("ftl: Failed to write back to BAM cache in copy_erase_unit()!\n");
		
		return ret;
	}


    /* Copy all data blocks from source unit to transfer unit */
    src = eun->Offset; dest = xfer->Offset;
    free = 0;
    ret = 0;
    for (i = 0; i < part->BlocksPerUnit; i++) {
	switch (BLOCK_TYPE(part->bam_cache[i])) {
	case BLOCK_CONTROL:
	    /* This gets updated later */
	    break;
	case BLOCK_DATA:
	case BLOCK_REPLACEMENT:
	    ret = MTD_READ(part->mtd, src, SECTOR_SIZE,
			&retlen, (u_char *) buf, 0);
	    if (ret)
	    {
		    printk("ftl: Error reading old xfer unit in copy_erase_unit\n");
		    return ret;
	    }
	    ret = MTD_WRITE(part->mtd, dest, SECTOR_SIZE,
			&retlen, (u_char *) buf, 0);
	    if (ret)
	    {
		    printk("ftl: Error writing new xfer unit in copy_erase_unit\n");
		    return ret;
	    }
	default:
	    /* All other blocks must be free */
	    part->bam_cache[i] = 0xffffffff;
	    free++;
	    break;
	}
	src += SECTOR_SIZE;
	dest += SECTOR_SIZE;
    }

    /* Write the BAM to the transfer unit */
    ret = MTD_WRITE(part->mtd, xfer->Offset + part->header.BAMOffset, 
		    SECTOR_SIZE, &retlen, (u_char *)part->bam_cache, 0);
    if (ret)
    {
	    printk("ftl: Error writing BAM in copy_erase_unit\n");
	    return ret;
    }    

    
    /* All clear? Then update the LogicalEUN again */
    ret = MTD_WRITE(part->mtd, xfer->Offset + 20, sizeof(u_short),
		    &retlen, (u_char *)&srcunit, 0);
    if (ret)
    {
	    printk("ftl: Error writing new LogicalEUN in copy_erase_unit\n");
	    return ret;
    }    


    /* Update the maps and usage stats*/
    i = xfer->EraseCount;
    xfer->EraseCount = eun->EraseCount;
    eun->EraseCount = i;
    i = xfer->Offset;
    xfer->Offset = eun->Offset;
    eun->Offset = i;
    part->FreeTotal -= eun->Free;
    part->FreeTotal += free;
/*    printk("Setting EUNInfo[%d].Free to %ld\n",srcunit,free);
 */  eun->Free = free;
    eun->Deleted = 0;
    
    /* Now, the cache should be valid for the new block */
    part->bam_index = srcunit;

    return 0;
    

} /* copy_erase_unit */

/*======================================================================

    reclaim_block() picks a full erase unit and a transfer unit and
    then calls copy_erase_unit() to copy one to the other.  Then, it
    schedules an erase on the expired block.

    What's a good way to decide which transfer unit and which erase
    unit to use?  Beats me.  My way is to always pick the transfer
    unit with the fewest erases, and usually pick the data unit with
    the most deleted blocks.  But with a small probability, pick the
    oldest data unit instead.  This means that we generally postpone
    the next reclaimation as long as possible, but shuffle static
    stuff around a bit for wear leveling.
    
======================================================================*/

static int reclaim_block(partition_t *part)
{
    u_short i, eun, xfer;
    u_long best;
    int queued, ret;

    DEBUG(0, "ftl_cs: reclaiming space...\n");
	
    /* Pick the least erased transfer unit */
    best = 0xffffffff; xfer = 0xffff;
    do {
	queued = 0;
	for (i = 0; i < part->header.NumTransferUnits; i++) {
	    if (part->XferInfo[i].state == XFER_UNKNOWN)
		erase_xfer(part, i);
	    if (part->XferInfo[i].state == XFER_ERASING)
		queued = 1;
	    else if (part->XferInfo[i].state == XFER_ERASED)
		prepare_xfer(part, i);
	    if ((part->XferInfo[i].state == XFER_PREPARED) &&
		(part->XferInfo[i].EraseCount <= best)) {
		    best = part->XferInfo[i].EraseCount;
		    xfer = i;
		}
	}
	if (xfer == 0xffff) {
	    if (queued) {
		DEBUG(1, "ftl_cs: waiting for transfer "
		      "unit to be prepared...\n");
		printk("Calling mtd_sync...");
		MTD_SYNC(part->mtd);
		printk(" Done\n");
	    }
	    else {
		static int ne = 0;
		if (++ne < 5)
		    printk(KERN_NOTICE "ftl_cs: reclaim failed: no "
			   "suitable transfer units!\n");
		return -EIO;
	    }
	}
    } while (xfer == 0xffff);
/*    printk("reclaim_block: decided on %d\n",xfer);
 */
    eun = 0;
    if ((jiffies % shuffle_freq) == 0) {
	DEBUG(1, "ftl_cs: recycling freshest block...\n");
	best = 0xffffffff;
	for (i = 0; i < part->DataUnits; i++)
	    if (part->EUNInfo[i].EraseCount <= best) {
		best = part->EUNInfo[i].EraseCount;
		eun = i;
	    }
    }
    else {
	best = 0;
	for (i = 0; i < part->DataUnits; i++)
	    if (part->EUNInfo[i].Deleted >= best) {
		best = part->EUNInfo[i].Deleted;
		eun = i;
	    }
	if (best == 0) {
	    static int ne = 0;
	    if (++ne < 5)
		printk(KERN_NOTICE "ftl_cs: reclaim failed: "
		       "no free blocks!\n");
	    return -EIO;
	}
    }
/*    printk("Calling copy_erase_unit (%p, %lx, %lx)\n",part,eun,xfer);
 */  
    ret = copy_erase_unit(part, eun, xfer);
    if (!ret)
	erase_xfer(part, xfer);
    else
	printk(KERN_NOTICE "ftl_cs: copy_erase_unit failed!\n");
    return ret;
} /* reclaim_block */

/*======================================================================

    Find_free() searches for a free block.  If necessary, it updates
    the BAM cache for the erase unit containing the free block.  It
    returns the block index -- the erase unit is just the currently
    cached unit.  If there are no free blocks, it returns 0 -- this
    is never a valid data block because it contains the header.
    
======================================================================*/

#ifdef PSYCHO_DEBUG
static void dump_lists(partition_t *part)
{
    int i;
    printk(KERN_DEBUG "ftl_cs: Free total = %ld\n", part->FreeTotal);
    for (i = 0; i < part->DataUnits; i++)
	printk(KERN_DEBUG "ftl_cs:   unit %d: %ld phys, %ld free, "
	       "%ld deleted\n", i,
	       part->EUNInfo[i].Offset >> part->header.EraseUnitSize,
	       part->EUNInfo[i].Free, part->EUNInfo[i].Deleted);
}
#endif

static inline u_long find_free(partition_t *part)
{
    u_short stop, eun;
    u_long blk, retlen;
    int ret;
    
    /* Find an erase unit with some free space */
    stop = (part->bam_index == 0xffff) ? 0 : part->bam_index;
 
    eun = stop;
    do {
	    if (part->EUNInfo[eun].Free != 0) break;
	    /* Wrap around at end of table */
	    if (++eun == part->DataUnits) eun = 0;
    } while (eun != stop);
    
    if (part->EUNInfo[eun].Free == 0)
	    return 0;
    
    /* Is this unit's BAM cached? */
    if (eun != part->bam_index) {
	    /* Invalidate cache */
	    part->bam_index = 0xffff;

	    ret = MTD_READ(part->mtd, 
			   part->EUNInfo[eun].Offset + part->header.BAMOffset,
			   part->BlocksPerUnit * sizeof(u_long),
			   &retlen, (u_char *) (part->bam_cache), 0);
	    
	    if (ret) {
		    printk("ftl: Error reading BAM in find_free\n");
		    return 0;
	    }

	    part->bam_index = eun;
    }
    
    /* Find a free block */
    for (blk = 0; blk < part->BlocksPerUnit; blk++)
	    if (BLOCK_FREE(part->bam_cache[blk])) break;
    
    if (blk == part->BlocksPerUnit) {
#ifdef PSYCHO_DEBUG
	    static int ne = 0;
	    if (++ne == 1)
		    dump_lists(part);
#endif
	    printk(KERN_NOTICE "ftl_cs: bad free list!\n");
	    return 0;
    }
    
    DEBUG(2, "ftl_cs: found free block at %ld in %d\n", blk, eun);
    return blk;
    
} /* find_free */

/*======================================================================

    This gets a memory handle for the region corresponding to the
    minor device number.
    
======================================================================*/



int ftl_open(struct inode *inode, struct file *file)
{
    int minor = MINOR(inode->i_rdev);
    partition_t *partition = &myparts[minor >> 4];

    

    printd("ftl_open, minor %d is device %d\n",minor, minor >>6);

    if (partition->state != FTL_FORMATTED)
    {
	    return -ENXIO;
    }

    if (ftl_gendisk.part[minor].nr_sects == 0)
    {
	    return -ENXIO;
    }


    if ((file->f_mode & 2) && !(partition->mtd->flags & MTD_CLEAR_BITS) )
	    return -EROFS;
    

    MTD_OPENCLOSE(partition->mtd,1);

    partition->open++;
  
    MOD_INC_USE_COUNT;
    return 0;
} /* ftl_open */

/*====================================================================*/

int ftl_close(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	partition_t *part = &myparts[minor >> 4];
	int i;

	printd("Now in ftl_close() minor %d device %d\n",minor,minor >> 4);
	

	/* Flush all writes */
	fsync_dev(inode->i_rdev);
/*	invalidate_inodes(inode->i_rdev);*/
	invalidate_buffers(inode->i_rdev);
	MTD_SYNC(part->mtd);

	
	printd("Flushed and invalidated\n");
	
	
        /* Wait for any pending erase operations to complete */
	 for (i = 0; i < part->header.NumTransferUnits; i++) 
	 {
		 if (part->XferInfo[i].state == XFER_ERASED)
			 prepare_xfer(part, i);
	 }
	   
	 
	 MTD_OPENCLOSE (part->mtd,0);
	 
	 printd("Ready to free stuff\n");
	 part->open--;
	 
	 
	 MOD_DEC_USE_COUNT;
	 return 0;
} /* ftl_close */



/*======================================================================

    Read a series of sectors from an FTL partition.
    
======================================================================*/

int ftl_read(partition_t *part, caddr_t buffer,
		    u_long sector, u_long nblocks)
{
    u_long log_addr, bsize, i, retlen;
    int ret; 
    printd("Entering ftl_read\n");

    DEBUG(2, "ftl_cs: ftl_read(0x%p, 0x%lx, %ld)\n",
	  part->handle, sector, nblocks);
    if (!(part->state & FTL_FORMATTED)) {
	printk(KERN_NOTICE "ftl_cs: bad partition\n");
	return -EIO;
    }
    bsize = part->mtd->erasesize;

    for (i = 0; i < nblocks; i++) {
	    if (((sector+i) * SECTOR_SIZE) >= part->header.FormattedSize) {
		    printk(KERN_NOTICE "ftl_cs: bad read offset\n");
		    return -EIO;
	    }
	    
/*      printd("sector is %lx, i is %lx\n",(u_long) sector, (u_long) i);*/
	    log_addr = part->VirtualBlockMap[sector+i];
/*	    printk("sector is %lu, logaddr %lu",sector+i,log_addr);*/
	    if (log_addr == 0xffffffff)
	    {
		    memset(buffer, 0, SECTOR_SIZE);
/*		    printk("\n");*/
	    }
	    else {
		    u_long offset = (part->EUNInfo[log_addr / bsize].Offset
				     + (log_addr % bsize));

		    ret = MTD_READ(part->mtd, offset, SECTOR_SIZE,
				   &retlen, (u_char *) buffer, 0);

		    if (ret)
		    {
			    printk("ftl: Failed to read device in ftl_read\n");
			    
			    return ret;
		    }
	    }
	    buffer += SECTOR_SIZE;
    }
    return 0;
} /* ftl_read */

/*======================================================================

    Write a series of sectors to an FTL partition
    
======================================================================*/

static inline int set_bam_entry(partition_t *part, u_long log_addr,
			 u_long virt_addr)
{
    u_long bsize, blk, offset, retlen;
#ifdef PSYCHO_DEBUG
    u_long old_addr;
#endif
    u_short eun;
    int ret;
    
    DEBUG(2, "ftl_cs: set_bam_entry(0x%p, 0x%lx, 0x%lx)\n",
	  part->handle, log_addr, virt_addr);
    bsize = part->mtd->erasesize;
    eun = log_addr / bsize;
    blk = (log_addr % bsize) / SECTOR_SIZE;

    offset = (part->EUNInfo[eun].Offset + blk * sizeof(u_long) +
		  part->header.BAMOffset);
    
#ifdef PSYCHO_DEBUG
    ret = MTD_READ(part->mtd, offset, sizeof(u_long),
			&retlen, (u_char *)&old_addr, 0);
	    if (ret)
	    {
		    printk("ftl: Error reading old_addr in set_bam_entry: %d\n",ret);
		    return ret;
		    
	    }
	    
    if (((virt_addr == 0xfffffffe) && !BLOCK_FREE(old_addr)) ||
	((virt_addr == 0) && (BLOCK_TYPE(old_addr) != BLOCK_DATA)) ||
	(!BLOCK_DELETED(virt_addr) && (old_addr != 0xfffffffe))) {
	static int ne = 0;
	if (1 || ++ne < 5) {
	    printk(KERN_NOTICE "ftl_cs: set_bam_entry() inconsistency!\n");
	    printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, old = 0x"
		   "%lx, new = 0x%lx\n", log_addr, old_addr, virt_addr);
	}
	printk("inconsistency(1)\n");
	/*	return -EIO;*/
    }
#endif
    if (part->bam_index == eun) {
#ifdef PSYCHO_DEBUG
	if (part->bam_cache[blk] != old_addr) {
	    static int ne = 0;
	    if (1 || ++ne < 5) {
		printk(KERN_NOTICE "ftl_cs: set_bam_entry() "
		       "inconsistency!\n");
		printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, cache"
		       " = 0x%lx, card = 0x%lx\n", log_addr,
		       part->bam_cache[blk], old_addr);
	    }
	    printk("inconsistency(2)!\n");
	    /*    return -EIO;*/
	}
#endif
	part->bam_cache[blk] = virt_addr;
    }
    ret = 0;

    ret = MTD_WRITE(part->mtd, offset, sizeof(u_long),
			    &retlen, (u_char *)&virt_addr, 0);
    if (ret)
    {
	    printk("ftl: Error writing virt_addr in set_bam_entry (%d)\n",ret);
#ifdef PSYCHO_DEBUG
	    printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, old = 0x%lx,"
		   " new = 0x%lx\n", log_addr, old_addr, virt_addr);
#endif
	    
    }    
    return ret;
    
} /* set_bam_entry */

static int ftl_write(partition_t *part, caddr_t buffer,
		     u_long sector, u_long nblocks)
{
	u_long bsize, log_addr, blk, i, offset, retlen;
	register u_long virt_addr;
	register u_long old_addr;
	register int ret;
	
	/* See if we need to reclaim space, before we start */
	while (part->FreeTotal < nblocks)
	{
		ret = reclaim_block(part);
		if (ret)
			return ret;
	}
	
	bsize = part->mtd->erasesize;
	
	virt_addr = sector * SECTOR_SIZE | BLOCK_DATA;
	for (i = 0; i < nblocks; i++)
	{
		if (virt_addr >= part->header.FormattedSize) 
			goto err_badsize;
		
		/* Grab a free block */
		blk = find_free(part);
		if (blk == 0)
		{
			static int ne = 0;
			if (++ne < 5)
				printk(KERN_NOTICE "ftl_cs: internal error: "
				       "no free blocks!\n");
			return -ENOSPC;
		}
		
		/* Tag the BAM entry, and write the new block */
		log_addr = part->bam_index * bsize + blk * SECTOR_SIZE;
		part->EUNInfo[part->bam_index].Free--;
		part->FreeTotal--;
		
		if ((ret = set_bam_entry(part, log_addr, 0xfffffffe)))
			goto err_bammark1;
		
		part->EUNInfo[part->bam_index].Deleted++;
		
		offset = (part->EUNInfo[part->bam_index].Offset +
			  blk * SECTOR_SIZE);
		
		if ((ret = MTD_WRITE(part->mtd, offset, SECTOR_SIZE, &retlen, 
				     buffer, 0)))
			goto err_writefail;
		
		/*printk("Write of user data succeeded at %lX and gave %lu bytes\n", offset, retlen);*/
		
		/* Only delete the old entry when the new entry is ready */
		
		old_addr = part->VirtualBlockMap[sector+i];
		if (old_addr != 0xffffffff) 
		{
			part->VirtualBlockMap[sector+i] = 0xffffffff;
			part->EUNInfo[old_addr/bsize].Deleted++;
			if ((ret = set_bam_entry(part, old_addr, 0)))
				goto err_bamdelete;
		}
		
		/* Finally, set up the new pointers */
		if ((ret=set_bam_entry(part, log_addr, virt_addr)))
			goto err_bamalloc;
		
		part->VirtualBlockMap[sector+i] = log_addr;
		part->EUNInfo[part->bam_index].Deleted--;
		
		buffer += SECTOR_SIZE;
		virt_addr += SECTOR_SIZE;
	}

	ret = 0;
out:
	return ret;
	
err_badsize:
	printk(KERN_NOTICE "ftl_cs: bad write offset\n");
	ret = -EIO;
	goto out;
	
err_bammark1:	
	printk("set_bam_entry failed to set bam entry to -2\n");
	goto out;
	
err_writefail:
	printk(KERN_NOTICE "ftl_cs: block write failed!\n");
	printk(KERN_NOTICE "ftl_cs:   log_addr = 0x%lx, virt_addr"
	       " = 0x%lx, Offset = 0x%lx\n", log_addr, virt_addr,
	       offset);
	goto out;

	
err_bamdelete:
	printk("set_bam_entry failed to delete old one\n");
	goto out;

	
err_bamalloc:
	printk("set_bam_entry failed to add new one\n");
	goto out;


} /* ftl_write */

/*======================================================================

    IOCTL calls for getting device parameters.

======================================================================*/


int ftl_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg)
{
    int minor = MINOR(inode->i_rdev);
    partition_t *partition=&myparts[minor >> 4];
    int ret = 0;


    printd("ftl_ioctl\n");

    switch (cmd) {
    case HDIO_GETGEO:
    {
	    struct hd_geometry loc;

	    loc.heads = 1;
	    loc.sectors = 8;
	    loc.cylinders = partition->header.FormattedSize / (SECTOR_SIZE * 8);
	    loc.start = ftl_gendisk.part[minor].start_sect;
	    if (copy_to_user((void *)arg, &loc, sizeof(loc)) < 0)
		    ret = -EFAULT;
	    break;
    }
    case BLKGETSIZE:
	    if (!(partition->state & FTL_FORMATTED)) return -EIO;
	    ret = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long));
	    if (!ret) 
		    put_user(ftl_gendisk.part[minor].nr_sects, (long *)arg);
	    break;
    case BLKFLSBUF:
	    if (!suser()) return -EACCES;
	    if (!(inode->i_rdev)) return -EINVAL;
	    fsync_dev(inode->i_rdev);
	    invalidate_buffers(inode->i_rdev);
	    break;
    case BLKRRPART:
	    for (ret=1; ret<16; ret++)
	    {
		    ftl_gendisk.part[minor+ret].nr_sects=0;
		    ftl_gendisk.part[minor+ret].start_sect=0;
	    }
	    resetup_one_dev(&ftl_gendisk,minor >> 4);
	    
	    
	    for (ret = 1; ret < 16; ret++) 
		    ftl_partsizes[minor + ret] = ftl_devsizes[minor + ret] =
			    ftl_gendisk.part[minor+ret].nr_sects >> (BLOCK_SIZE_BITS - 9);
	    ret = 0;
	    break;
	    
    default:
	    ret = -EINVAL;
    }
    printd("ftl_ioctl done\n");
    return ret;
} /* ftl_ioctl */

/*======================================================================

    Handler for block device requests

======================================================================*/

void do_ftl_request(void)
{
    int ret, minor;
    partition_t *part;

    printd("do_ftl_request\n");

    do {
	    sti();
	    INIT_REQUEST;
	    
	    minor = MINOR(CURRENT->rq_dev);
	    part=&myparts[minor >> 4];  
	    printd("minor %d device %d\n",minor , minor >> 4);
	    if (!part) 
	    {
		    printk("do_ftl_request: device %d doesn't exist\n",minor);
		    end_request(1);
	    }
	    if (part->state != FTL_FORMATTED) 
	    {
		    printk("do_ftl_request: device %d hasn't got FTL on it\n",minor);
		    end_request(1);
	    }
	    
	    ret = 0;
	    
	    
	    switch (CURRENT->cmd) {
		    
	    case READ:
		    printd("Calling ftl_read for %ld sectors from %ld\n",CURRENT->current_nr_sectors, CURRENT->sector);
		    ret = ftl_read(part, CURRENT->buffer, 
				   CURRENT->sector+ftl_gendisk.part[minor].start_sect,
				   CURRENT->current_nr_sectors);
		    break;
		    
	    case WRITE:
		    ret = ftl_write(part, CURRENT->buffer, 
				    CURRENT->sector + ftl_gendisk.part[minor].start_sect,
				    CURRENT->current_nr_sectors);
		    break;
		    
	    default:
		    panic("ftl_cs: unknown block command %d!\n", CURRENT->cmd);
		    
	    }
	    if (ret == 0)
		    end_request(1);
	    else {
		    static int ne = 0;
		    if (++ne < 5)
			    end_request(0);
		    else
			    end_request(1);
	    }
    } while (1);
} /* do_ftl_request */

/*====================================================================*/


static void ftl_notify_func(struct mtd_info *mtd, int device)
{
	partition_t *partition = &myparts[device];
	int i;

	if (!mtd)
	{
		/* Removing device */
		if (partition->state == FTL_FORMATTED)
			ftl_freepart(partition);
		partition->state = 0;
		for (i=0; i<16; i++)
		{
			ftl_devsizes[device << 4] = 0;
			ftl_partsizes[device << 4] = 0;
			ftl_gendisk.part[i].nr_sects=0;
			ftl_gendisk.part[i].start_sect=0;
		}
	}
	else
	{
		/* Adding device */
		if (partition->state == FTL_FORMATTED)
			return;
		partition->state = 0;
		partition->mtd   = mtd;
		if (!scan_partition(partition))
		{
			partition->state = FTL_FORMATTED;
			printk(KERN_INFO "ftl_cs: opening %ld kb FTL partition\n",
			       partition->header.FormattedSize >> 10);
			ftl_partsizes[device << 4] = ftl_devsizes[device << 4] = 
				partition->header.FormattedSize >> 10;

			printd("Calling setup_dev...\n");
			
			ftl_gendisk.part[device << 4].nr_sects =
				partition->header.FormattedSize >> 9;
			
			
			resetup_one_dev(&ftl_gendisk,device);
	

			for (i = 1; i < 16; i++) 
				ftl_partsizes[(device << 4) + i] = ftl_devsizes[(device << 4) + i] =
					ftl_gendisk.part[(device << 4) + i].nr_sects >> (BLOCK_SIZE_BITS - 9);
		}
		else
			ftl_freepart(&myparts[device]);
	}
}

void ftl_freepart(partition_t *part)
{
	if (part->VirtualBlockMap) {
		vfree(part->VirtualBlockMap);
		part->VirtualBlockMap = NULL;
	}
	if (part->VirtualPageMap) {
		kfree(part->VirtualPageMap);
		part->VirtualPageMap = NULL;
	}
	if (part->EUNInfo) {
		kfree(part->EUNInfo);
		part->EUNInfo = NULL;
	}
	if (part->XferInfo) {
		kfree(part->XferInfo);
		part->XferInfo = NULL;
	}
	if (part->bam_cache) {
		kfree(part->bam_cache);
		part->bam_cache = NULL;
	}
}

int init_module(void)
{
	int i;
	struct mtd_info *mtd;
	
	memset(&myparts[0],0,sizeof(myparts));
	
	if (register_blkdev(FTL_MAJOR,"ftl", &ftl_blk_fops))
	{
		printk(KERN_NOTICE "ftl_cs: unable to grab major "
		       "device number %d!\n",FTL_MAJOR);
		return EAGAIN;
	}

	blk_dev[FTL_MAJOR].request_fn = DEVICE_REQUEST;
	
	for (i = 0; i < 256; i++)
	{
		ftl_blocksizes[i] = 1024;	
		ftl_devsizes[i] = 0;
		ftl_partsizes[i] = 0;
		ftl_gendisk.part[i].start_sect = 0;
		ftl_gendisk.part[i].nr_sects = 0;
	}
	


	blksize_size[FTL_MAJOR]=ftl_blocksizes;
	blk_size[FTL_MAJOR] = ftl_devsizes;
	
	register_mtd_notifier(&ftl_notifier);

	for (i=0; i< MAX_MTD_DEVICES; i++)
	{
		mtd = get_mtd_device(i);
		if (!mtd)
			continue;
		
		ftl_notify_func(mtd, i);
	}		    


	return 0;

}

void cleanup_module(void)
{
	int i;

	unregister_mtd_notifier(&ftl_notifier);
	for (i=0; i < MAX_MTD_DEVICES; i++)
		ftl_notify_func(NULL, i);

	
	unregister_blkdev(FTL_MAJOR, "ftl");
}

