/*======================================================================

  $Id: mtd.c,v 1.2 1999/01/08 16:54:45 dwmw2 Exp $

    A general driver for accessing PCMCIA card memory via Bulk
    Memory Services.

    This driver provides the equivalent of /dev/mem for a PCMCIA
    card's attribute and common memory.  It includes character
    and block devices.

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <stdarg.h>

#include "mtd.h"

/* Major device #'s for memory device */

static struct mtd_info *mtd_table[MAX_MTD_DEVICES];



static loff_t mtd_lseek (struct file *file, loff_t offset, int orig);

static int mtd_ioctl(struct inode *inode, struct file *file,
		     u_int cmd, u_long arg);
static ssize_t mtd_read(struct file *file, char *buf, size_t count,
		     loff_t *ppos);
static ssize_t mtd_write(struct file *file, const char *buf, size_t count,
			 loff_t *ppos);
static int mtd_open(struct inode *inode, struct file *file);
static int mtd_close(struct inode *inode,
		     struct file *file);

static struct file_operations mtd_fops = {
	mtd_lseek,     	/* lseek */
	mtd_read,	/* read */
	mtd_write, 	/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
	mtd_ioctl,	/* ioctl */
	NULL,		/* mmap */
	mtd_open,	/* open */
	NULL,		/* flush */
	mtd_close,	/* release */
	NULL		/* fsync */
};

static struct mtd_notifier *mtd_notifiers = NULL;

static loff_t mtd_lseek (struct file *file, loff_t offset, int orig)
{
	int devnum = MINOR(file->f_dentry->d_inode->i_rdev) >> 1;
	struct mtd_info *mtd=mtd_table[devnum];

	if (devnum >=MAX_MTD_DEVICES || !mtd) 
		return -ENODEV;

	switch (orig) {
	case 0: 
		/* SEEK_SET */
		file->f_pos = offset;
		break;
	case 1: 
		/* SEEK_CUR */
		file->f_pos += offset;
		break;
	case 2:
		/* SEEK_END */
		file->f_pos =mtd->size + offset;
		break;
	default: 
		return -EINVAL;
	}

	if (file->f_pos < 0) 
		file->f_pos = 0;
	else if (file->f_pos >= mtd->size)
		file->f_pos = mtd->size - 1;

	return file->f_pos;
}

static int mtd_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	int devnum = minor >> 1;
	struct mtd_info *mtd=mtd_table[devnum];	
	
	if (devnum >= MAX_MTD_DEVICES || !mtd)
		return -ENODEV;
	
	if (file->f_mode & 2)
	{
		/* We're trying to open it with write access. */
		/* Fail if it's read-only, or if we're using the ro device */
		if ((minor & 1) || !(mtd->flags & MTD_WRITEABLE))
			return -EACCES;
	}

	/* OK. We're going to open it now. */

	mtd->usage++;	
	(*(mtd->openclose))(mtd, 1);

	return 0;
} /* mtd_open */

/*====================================================================*/

static int mtd_close(struct inode *inode,
				 struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct mtd_info *mtd;
	int devnum = minor >> 1;

	if (devnum >= MAX_MTD_DEVICES)
		return -ENODEV;

	mtd = mtd_table[devnum];
	
	if (!mtd) return -ENODEV;
	MTD_SYNC(mtd);
	(*(mtd->openclose))(mtd,0);
	mtd->usage--;	  

	return 0;
} /* mtd_close */


static ssize_t mtd_read(struct file *file, char *buf, size_t count,
			loff_t *ppos)
{
	int devnum = MINOR(file->f_dentry->d_inode->i_rdev) >> 1;
	struct mtd_info *mtd = mtd_table[devnum];
	u_char *mtdbuf;
	u_long retlen=0;
	u_long count1;
	int ret=0;

	if (!mtd) return -ENODEV;

	if (*ppos + count > mtd->size)
		count = mtd->size - *ppos;

	if (!count)
		printk("Count is zero.\n");

	count1 = count;

	while (count && !ret)
	{
		ret = MTD_POINT(mtd, *ppos, count, &retlen, &mtdbuf);
		if (ret)
		{
			/* mtd->point() failed; use mtd->read instead */
		
			ret = MTD_READ(mtd, *ppos, count, &retlen, buf, 1);
			if (retlen)
			{
				*ppos += retlen;
				return retlen;
			}
			else 
				return ret;
		}

		if (!retlen)
		{
			printk(KERN_NOTICE "mtd_read: device returned no error, but no data either!\n");
			ret = -EIO;
			break;
		}
		
		if (copy_to_user(buf,mtdbuf,retlen) < 0) 
			ret = -EFAULT;
		else
		{
			count -= retlen;
			*ppos += retlen;
		}
		MTD_UNPOINT(mtd, *ppos);
	}
	
	if (count < count1)
		return count1-count;
	else 
		return ret;
} /* mtd_read */


static ssize_t mtd_write(struct file *file, const char *buf, size_t count,
			 loff_t *ppos)
{
	int devnum = MINOR(file->f_dentry->d_inode->i_rdev) >> 1;
	struct mtd_info *mtd = mtd_table[devnum];
	u_long count1;
	u_long retlen;
	int ret=0;
	
	if (!mtd) return -ENODEV;

	if (*ppos == mtd->size)
		return -ENOSPC;
	
	if (*ppos + count > mtd->size)
		count = mtd->size - *ppos;
	
	count1 = count;
	
	while (count)
	{
		ret = (*(mtd->write))(mtd, *ppos, count, &retlen, buf, 1);

		if (ret)
			break;

		if (!retlen)
		{
			printk(KERN_NOTICE "mtd_read: device returned no error, but no data either!\n");
			ret = -EIO;
			break;
		}

		count -= retlen;
	}

	if (count < count1)
	{
		*ppos += count1-count;
		return count1-count;
	}
	else 
		return ret;
} /* mtd_write */

/*======================================================================

    IOCTL calls for getting device parameters.

======================================================================*/

static int mtd_ioctl(struct inode *inode, struct file *file,
		       u_int cmd, u_long arg)
{
	int devnum = MINOR(inode->i_rdev) >> 1;
	struct mtd_info *mtd=mtd_table[devnum];
	int ret = 0;
	u_long size;
	

	if (devnum >= MAX_MTD_DEVICES || !mtd)
		return -ENODEV;
	
	size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
	if (cmd & IOC_IN) {
		ret = verify_area(VERIFY_READ, (char *)arg, size);
		if (ret) return ret;
	}
	if (cmd & IOC_OUT) {
		ret = verify_area(VERIFY_WRITE, (char *)arg, size);
		if (ret) return ret;
	}
	
	switch (cmd) {
	case MEMGETINFO:
		copy_to_user((struct mtd_info *)arg, mtd,
			     (u_long) &mtd->erase - (u_long)mtd);
		break;
	case MEMERASE:
	{
		struct erase_info *erase=kmalloc(sizeof(struct erase_info),GFP_KERNEL);
		if (!erase)
			ret = -ENOMEM;
		else
		{
			memset (erase,0,sizeof(struct erase_info));
			copy_from_user(&erase->addr, (u_long *)arg,
				       2 * sizeof(u_long));
			erase->mtd = mtd;
			ret = MTD_ERASE(mtd, erase);
		}
		break;
	}
	default:
		ret = -EINVAL;
	}
	
	return ret;
} /* memory_ioctl */

/*====================================================================*/

int add_mtd_device(struct mtd_info *mtd)
{
	int i;

	for (i=0; i< MAX_MTD_DEVICES; i++)
		if (!mtd_table[i])
		{
			struct mtd_notifier *not=mtd_notifiers;

			mtd_table[i]=mtd;
			
			printk(KERN_NOTICE "mtd: Giving out device %d to %s\n",i, mtd->name);
			while (not)
			{
				(*(not->func))(mtd,i);
				not = not->next;
			}
			MOD_INC_USE_COUNT;
			return 0;
		}
	
	return 1;
}

struct mtd_info *get_mtd_device(u_char num)
{
	return mtd_table[num];
}

int del_mtd_device (struct mtd_info *mtd)
{
	struct mtd_notifier *not=mtd_notifiers;
	int i;
	
	if (mtd->usage)
	    return -EAGAIN;

	for (i=0; i < MAX_MTD_DEVICES; i++)
	{
		if (mtd_table[i] == mtd)
		{
			while (not)
			{
				(*(not->func))(NULL,i);
				not=not->next;
			}
			mtd_table[i] = NULL;
			MOD_DEC_USE_COUNT;

			return 0;
		}
	}

	return 1;
}



void register_mtd_notifier (struct mtd_notifier *new)
{
	new->next = mtd_notifiers;
	mtd_notifiers = new;
	MOD_INC_USE_COUNT;
}



int unregister_mtd_notifier (struct mtd_notifier *old)
{
	struct mtd_notifier **prev = &mtd_notifiers;
	struct mtd_notifier *cur;

	while ((cur = *prev))
	{
		if (cur == old)
		{
			*prev = cur->next;
			MOD_DEC_USE_COUNT;
			return 0;
		}
		prev = &cur->next;
	}
	return 1;
}
	

/*====================================================================*/

int init_module(void)
{
	int i;
	
	for (i=0; i<MAX_MTD_DEVICES; i++)
		mtd_table[i]=NULL;

	if (register_chrdev(MTD_MAJOR,"mtd",&mtd_fops))
	{
		printk(KERN_NOTICE "Can't allocate major number %d for Memory Technology Devices.\n",
		       MTD_MAJOR);
		return EAGAIN;
	}
	return 0;
}

void cleanup_module(void)
{
	unregister_chrdev(MTD_MAJOR, "mtd");
}
